"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const languageConfigurations_1 = require("./languageConfigurations");
const xpathDocumentChangeHandler_1 = require("./xpathDocumentChangeHandler");
const xslLexerLight_1 = require("./xslLexerLight");
const xslLexerRenameTag_1 = require("./xslLexerRenameTag");
class DocumentChangeHandler {
    constructor() {
        this.onDidChangeRegistration = null;
        this.xmlDocumentRegistered = false;
        this.lastChangePerformed = null;
        this.lexer = new xslLexerRenameTag_1.XslLexerRenameTag(languageConfigurations_1.XMLConfiguration.configuration);
        this.cachedFailedEdit = null;
        this.xpathDocumentChangeHanlder = null;
        this.registerXMLEditor = (editor) => {
            if (editor) {
                this.registerXMLDocument(editor);
            }
            else {
            }
        };
        this.registerXMLDocument = (editor) => {
            const document = editor.document;
            let isXMLDocument = document.languageId === 'xml' || document.languageId === 'xslt' || document.languageId === 'dcp' || document.languageId === 'sch';
            let isXPathDocument = document.languageId === 'xpath';
            if (this.xmlDocumentRegistered && (!isXMLDocument || !isXPathDocument) && this.onDidChangeRegistration) {
                this.onDidChangeRegistration.dispose();
                this.xmlDocumentRegistered = false;
            }
            if (isXMLDocument) {
                DocumentChangeHandler.lastActiveXMLEditor = editor;
                DocumentChangeHandler.getLastDocXmlnsPrefixes();
            }
            if (isXMLDocument && !this.xmlDocumentRegistered) {
                this.xmlDocumentRegistered = true;
                this.onDidChangeRegistration = vscode.workspace.onDidChangeTextDocument(e => this.onDocumentChange(e, isXMLDocument));
            }
            else if (isXPathDocument && !this.xmlDocumentRegistered) {
                this.xmlDocumentRegistered = true;
                this.onDidChangeRegistration = vscode.workspace.onDidChangeTextDocument(e => this.getXPathDocumentChangeHandler().onDocumentChange(e));
            }
        };
    }
    onDocumentChange(e, isXML) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!isXML) {
                return;
            }
            let activeChange = e.contentChanges[0];
            let skipTrigger = false;
            if (!activeChange) {
                return;
            }
            let triggerSuggest = false;
            if (activeChange.text === '/' && activeChange.rangeLength === 0) {
                let nextChar = e.document.getText().charAt(activeChange.rangeOffset + 1);
                if (nextChar === '>') {
                    let textBeforeCloseEnd = this.lexer.getTextBeforeCloseEnd(e.document, activeChange);
                    if (textBeforeCloseEnd) {
                        let newStartChar = (activeChange.range.start.character - textBeforeCloseEnd.length); // positive number
                        let activeStartPos = activeChange.range.start;
                        let newStartPos = new vscode.Position(activeStartPos.line, newStartChar + 1);
                        let newRange = new vscode.Range(newStartPos, newStartPos);
                        let newOffset = activeChange.rangeOffset - (textBeforeCloseEnd.length + 1);
                        let endTagPosData = this.lexer.getEndTagForStartTagChange(e.document, newOffset, activeChange.range.start.line, newStartChar, newRange, true);
                        if (endTagPosData && (endTagPosData === null || endTagPosData === void 0 ? void 0 : endTagPosData.startTag.endsWith('/'))) {
                            endTagPosData.startTag = endTagPosData.startTag.substring(0, endTagPosData.startTag.length - 1);
                        }
                        if (endTagPosData && endTagPosData.startTag === endTagPosData.endTag) {
                            let endLine = e.document.lineAt(endTagPosData.startPosition.line).text;
                            let followingStartCloseTag = endLine.substring(endTagPosData.startPosition.character);
                            let closeTagCharPos = followingStartCloseTag.indexOf('>');
                            if (closeTagCharPos > -1) {
                                let startPos = new vscode.Position(activeChange.range.start.line, activeChange.range.start.character + 1);
                                let endCharPos = endTagPosData.startPosition.character + closeTagCharPos;
                                let endPos = new vscode.Position(endTagPosData.startPosition.line, endCharPos);
                                let deletionRange = new vscode.Range(startPos, endPos);
                                let wse = new vscode.WorkspaceEdit();
                                wse.delete(e.document.uri, deletionRange);
                                vscode.workspace.applyEdit(wse);
                            }
                        }
                    }
                }
                else {
                    let prevChar = e.document.getText().charAt(activeChange.rangeOffset - 1);
                    if (prevChar === '"' || activeChange.text.endsWith('>') || activeChange.text.endsWith('/')) {
                        skipTrigger = true;
                    }
                }
            }
            if (!skipTrigger && activeChange.rangeOffset > 10) {
                let prevChar = e.document.getText().charAt(activeChange.rangeOffset - 1);
                if ((prevChar === '"' || prevChar === '(') && activeChange.text.length === 1 && ['[', '(', '{', '?', '"', '\''].indexOf(activeChange.text) === -1) {
                    triggerSuggest = true;
                }
            }
            // console.log('activeChange.text:', activeChange.text, 'triggerSuggest', triggerSuggest);
            if (triggerSuggest || activeChange.text === '(' || activeChange.text === '/' || activeChange.text === '[' || activeChange.text === '!' || activeChange.text === '$' || activeChange.text === '<') {
                // console.log('triggering...');
                setTimeout(() => {
                    vscode.commands.executeCommand('editor.action.triggerSuggest');
                }, 10);
                return;
            }
            if (this.lastChangePerformed === null || !this.changesAreEqual(this.lastChangePerformed, activeChange)) {
                let tagNameLengthBeforeEdit = this.lexer.isStartTagChange(e.document, activeChange);
                if (tagNameLengthBeforeEdit > -1) {
                    let startBracketPos = activeChange.range.start.character - (tagNameLengthBeforeEdit + 1);
                    let startTagOffset = activeChange.range.start.character - startBracketPos; // positive number
                    let offset = activeChange.rangeOffset - startTagOffset;
                    let line = activeChange.range.start.line;
                    let character = startBracketPos;
                    let endTagPosData = this.lexer.getEndTagForStartTagChange(e.document, offset, line, character, activeChange.range, false);
                    if (endTagPosData) {
                        if (this.cachedFailedEdit) {
                            this.cachedFailedEdit = null;
                            let startChar = endTagPosData.startPosition.character;
                            let startLline = endTagPosData.startPosition.line;
                            let updateStartPos = new vscode.Position(startLline, startChar);
                            let updateEndPos = new vscode.Position(startLline, startChar + endTagPosData.endTag.length);
                            let updateRange = new vscode.Range(updateStartPos, updateEndPos);
                            // replace the whole endtag to be safe:
                            this.lastChangePerformed = { range: updateRange, text: endTagPosData.startTag, fullTagName: endTagPosData.startTag };
                            yield this.performRename(e.document, Object.assign(this.lastChangePerformed));
                        }
                        let endTagNameOk = this.checkEndTag(endTagPosData, tagNameLengthBeforeEdit, activeChange);
                        if (endTagNameOk) {
                            let endTagPos = endTagPosData.startPosition;
                            let adjustedStartTagPos = endTagPos.character + (tagNameLengthBeforeEdit);
                            let updateStartPos = new vscode.Position(endTagPos.line, adjustedStartTagPos);
                            let updateEndPos = new vscode.Position(endTagPos.line, adjustedStartTagPos + activeChange.rangeLength);
                            let updateRange = new vscode.Range(updateStartPos, updateEndPos);
                            this.lastChangePerformed = { range: updateRange, text: activeChange.text, fullTagName: endTagPosData.startTag };
                            yield this.performRename(e.document, Object.assign(this.lastChangePerformed));
                        }
                        else {
                            this.lastChangePerformed = null;
                        }
                    }
                }
                else {
                    this.lastChangePerformed = null;
                }
            }
            else {
                this.lastChangePerformed = null;
            }
        });
    }
    checkEndTag(endTagData, startTagPos, startTagChange) {
        let result = false;
        let endTagName = endTagData.endTag;
        if (startTagPos > endTagName.length) {
            return false;
        }
        let originalStartTagLength = endTagData.startTag.length - (startTagChange.text.length - startTagChange.rangeLength);
        if (originalStartTagLength !== endTagData.endTag.length) {
            return false;
        }
        let beforeEndTag = endTagName.substring(0, startTagPos);
        let afterEndTag = endTagName.substring(startTagPos + startTagChange.rangeLength);
        let updatedEndName = beforeEndTag + startTagChange.text + afterEndTag;
        result = updatedEndName === endTagData.startTag;
        return result;
    }
    getXPathDocumentChangeHandler() {
        if (this.xpathDocumentChangeHanlder === null) {
            this.xpathDocumentChangeHanlder = new xpathDocumentChangeHandler_1.XPathDocumentChangeHandler();
        }
        return this.xpathDocumentChangeHanlder;
    }
    performRename(document, edit) {
        return __awaiter(this, void 0, void 0, function* () {
            let wse = new vscode.WorkspaceEdit();
            wse.replace(document.uri, edit.range, edit.text);
            let success = false;
            yield vscode.workspace.applyEdit(wse).then((result) => {
                success = result;
            });
            if (success) {
                this.cachedFailedEdit = null;
            }
            else {
                this.cachedFailedEdit = edit;
                this.lastChangePerformed = null;
            }
            return success;
        });
    }
    changesAreEqual(tagRenameEdit, change2) {
        if (tagRenameEdit && change2) {
            let result = (tagRenameEdit.range.start.line === change2.range.start.line &&
                tagRenameEdit.range.start.character === change2.range.start.character &&
                tagRenameEdit.range.end.character === change2.range.end.character &&
                tagRenameEdit.range.end.line === change2.range.end.line &&
                tagRenameEdit.text === change2.text);
            return result;
        }
        else {
            return false;
        }
    }
    static getLastDocXmlnsPrefixes() {
        if (DocumentChangeHandler.lastActiveXMLEditor) {
            DocumentChangeHandler.lastXMLDocumentGlobalData = this.lexer.analyseLight(DocumentChangeHandler.lastActiveXMLEditor.document.getText(), true);
        }
    }
}
exports.DocumentChangeHandler = DocumentChangeHandler;
DocumentChangeHandler.lastActiveXMLEditor = null;
DocumentChangeHandler.lastXMLDocumentGlobalData = [];
DocumentChangeHandler.lexer = new xslLexerLight_1.XslLexerLight(languageConfigurations_1.XSLTLightConfiguration.configuration);
//# sourceMappingURL=documentChangeHandler.js.map