"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable
const xslLexerLight_1 = require("./xslLexerLight");
const xslLexer_1 = require("./xslLexer");
const vscode = require("vscode");
const xsltSymbolProvider_1 = require("./xsltSymbolProvider");
const url = require("url");
class DocumentLinkProvider {
    constructor(languageConfig) {
        this.lexer = new xslLexerLight_1.XslLexerLight(languageConfig);
    }
    provideDocumentLinks(document, token) {
        let data = this.lexer.analyseLight(document.getText());
        let result = [];
        const xsltPackages = vscode.workspace.getConfiguration('XSLT.resources').get('xsltPackages');
        const rootPath = vscode.workspace.rootPath;
        data.forEach((instruction) => {
            if (instruction.type === xslLexer_1.GlobalInstructionType.Import || instruction.type === xslLexer_1.GlobalInstructionType.Include) {
                const resolvedPath = xsltSymbolProvider_1.XsltSymbolProvider.resolvePath(instruction.name, document.fileName);
                const pathForUri = resolvedPath.startsWith('file:/') ? resolvedPath : url.pathToFileURL(resolvedPath).toString();
                const uri = vscode.Uri.parse(pathForUri);
                const startPos = new vscode.Position(instruction.token.line, instruction.token.startCharacter);
                const endPos = new vscode.Position(instruction.token.line, instruction.token.startCharacter + (instruction.token.length + 2));
                const link = new vscode.DocumentLink(new vscode.Range(startPos, endPos), uri);
                result.push(link);
            }
            else if (instruction.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                let packageLookup = xsltPackages.find((pkg) => {
                    return pkg.name === instruction.name;
                });
                if (packageLookup && rootPath) {
                    let resolvedName = xsltSymbolProvider_1.XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                    const pathForUri = resolvedName.startsWith('file:/') ? resolvedName : url.pathToFileURL(resolvedName).toString();
                    const uri = vscode.Uri.parse(pathForUri);
                    const startPos = new vscode.Position(instruction.token.line, instruction.token.startCharacter);
                    const endPos = new vscode.Position(instruction.token.line, instruction.token.startCharacter + (instruction.token.length + 2));
                    const link = new vscode.DocumentLink(new vscode.Range(startPos, endPos), uri);
                    result.push(link);
                }
            }
        });
        return result;
    }
}
exports.DocumentLinkProvider = DocumentLinkProvider;
//# sourceMappingURL=documentLinkProvider.js.map