"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const xslLexer_1 = require("./xslLexer");
const xsltTokenDiagnostics_1 = require("./xsltTokenDiagnostics");
const globalsProvider_1 = require("./globalsProvider");
const path = require("path");
class XsltSymbolProvider {
    constructor(xsltConfiguration, collection) {
        this.gp = new globalsProvider_1.GlobalsProvider();
        this.isXSLT = xsltConfiguration.nativePrefix === 'xsl';
        this.xslLexer = new xslLexer_1.XslLexer(xsltConfiguration);
        this.xslLexer.provideCharLevelState = true;
        this.collection = collection;
    }
    provideDocumentSymbols(document, token) {
        return __awaiter(this, void 0, void 0, function* () {
            const allTokens = this.xslLexer.analyse(document.getText());
            const globalInstructionData = this.xslLexer.globalInstructionData;
            // Import/include XSLT - ensuring no duplicates
            let importedG = { data: globalInstructionData, href: document.fileName, error: false };
            let importedGlobals1 = [importedG];
            let accumulatedHrefs = [importedG.href];
            let topLevelHrefs = this.accumulateImportHrefs(importedGlobals1, []);
            let globalsSummary0 = { globals: importedGlobals1, hrefs: accumulatedHrefs };
            const maxImportLevel = 20;
            let processNestedGlobals = () => __awaiter(this, void 0, void 0, function* () {
                let level = 0;
                while (globalsSummary0.hrefs.length > 0 && level < maxImportLevel) {
                    globalsSummary0 = yield this.processImportedGlobals(globalsSummary0.globals, accumulatedHrefs, level === 0);
                    level++;
                }
            });
            yield processNestedGlobals();
            return new Promise((resolve, reject) => {
                let symbols = [];
                let allImportedGlobals = [];
                let importErrors = [];
                globalsSummary0.globals.forEach((globals) => {
                    if (globals.error) {
                        if (topLevelHrefs.indexOf(globals.href) > -1) {
                            let errorData = globalInstructionData.find((dataObject) => {
                                let result = false;
                                if (dataObject.type === xslLexer_1.GlobalInstructionType.Import || dataObject.type === xslLexer_1.GlobalInstructionType.Include) {
                                    let resolvedName = this.resolvePath(dataObject.name, document.fileName);
                                    result = resolvedName === globals.href;
                                }
                                return result;
                            });
                            if (errorData) {
                                importErrors.push(errorData);
                            }
                        }
                    }
                    else {
                        globals.data.forEach((global) => {
                            global['href'] = globals.href;
                            allImportedGlobals.push(global);
                        });
                    }
                });
                let importDiagnostics = [];
                importErrors.forEach((importError) => {
                    importDiagnostics.push(xsltTokenDiagnostics_1.XsltTokenDiagnostics.createImportDiagnostic(importError));
                });
                let diagnostics = xsltTokenDiagnostics_1.XsltTokenDiagnostics.calculateDiagnostics(this.isXSLT, document, allTokens, globalInstructionData, allImportedGlobals, symbols);
                let allDiagnostics = importDiagnostics.concat(diagnostics);
                if (allDiagnostics.length > 0) {
                    this.collection.set(document.uri, allDiagnostics);
                }
                else {
                    this.collection.clear();
                }
                ;
                resolve(symbols);
            });
        });
    }
    processImportedGlobals(importedGlobals1, level1Hrefs, topLevel) {
        return __awaiter(this, void 0, void 0, function* () {
            let level2Globals = [];
            let level2Hrefs = this.accumulateImportHrefs(importedGlobals1, level1Hrefs);
            let newGlobals = [];
            level2Hrefs.forEach((href) => {
                level2Globals.push(this.fetchImportedGlobals([href]));
            });
            let importedGlobals2Array = yield Promise.all(level2Globals);
            importedGlobals2Array.forEach((importedGlobals2) => {
                if (topLevel) {
                    newGlobals = newGlobals.concat(importedGlobals2);
                }
                else {
                    importedGlobals1 = importedGlobals1.concat(importedGlobals2);
                }
            });
            if (topLevel) {
                return { globals: newGlobals, hrefs: level2Hrefs };
            }
            else {
                return { globals: importedGlobals1, hrefs: level2Hrefs };
            }
        });
    }
    accumulateImportHrefs(importedGlobals, existingHrefs) {
        let result = [];
        importedGlobals.forEach((importedG) => {
            importedG.data.forEach((data) => {
                if (data.type === xslLexer_1.GlobalInstructionType.Import || data.type === xslLexer_1.GlobalInstructionType.Include) {
                    let resolvedName = this.resolvePath(data.name, importedG.href);
                    if (existingHrefs.indexOf(resolvedName) < 0) {
                        existingHrefs.push(resolvedName);
                        result.push(resolvedName);
                    }
                }
            });
        });
        return result;
    }
    resolvePath(href, documentPath) {
        if (path.isAbsolute(href)) {
            return href;
        }
        else {
            let basePath = path.dirname(documentPath);
            let joinedPath = path.join(basePath, href);
            return path.normalize(joinedPath);
        }
    }
    fetchImportedGlobals(inputHrefs) {
        return __awaiter(this, void 0, void 0, function* () {
            let result = [];
            //let inputHrefs: string[] = this.accumulateImportHrefs(globalInstructionData, existingHrefs, docHref);
            let lastIndex = inputHrefs.length - 1;
            if (lastIndex < 0) {
                return result;
            }
            else {
                return new Promise((resolve, reject) => {
                    inputHrefs.forEach((href, index) => {
                        this.gp.provideGlobals(href).then((globals) => {
                            result.push({ href: href, data: globals.data, error: globals.error });
                            if (index === lastIndex) {
                                resolve(result);
                            }
                        });
                    });
                });
            }
        });
    }
}
exports.XsltSymbolProvider = XsltSymbolProvider;
//# sourceMappingURL=xsltSymbolProvider copy.js.map